// ProfileFactorial.c
// Runs on TM4C1294
// Daniel Valvano
// May 6, 2014

/* This example accompanies the books
   "Embedded Systems: Introduction to ARM Cortex M Microcontrollers",
   ISBN: 978-1469998749, Jonathan Valvano, copyright (c) 2014
   Volume 1, Program 5.20

 Copyright 2014 by Jonathan W. Valvano, valvano@mail.utexas.edu
    You may use, edit, run or distribute this file
    as long as the above copyright notice remains
 THIS SOFTWARE IS PROVIDED "AS IS".  NO WARRANTIES, WHETHER EXPRESS, IMPLIED
 OR STATUTORY, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF
 MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE APPLY TO THIS SOFTWARE.
 VALVANO SHALL NOT, IN ANY CIRCUMSTANCES, BE LIABLE FOR SPECIAL, INCIDENTAL,
 OR CONSEQUENTIAL DAMAGES, FOR ANY REASON WHATSOEVER.
 For more information about my classes, my research, and my books, see
 http://users.ece.utexas.edu/~valvano/
 */

// PN2-1 is an output for debugging

#include <stdint.h>
#include "PLL.h"

#define GPIO_PORTN21            (*((volatile uint32_t *)0x40064018))
#define GPIO_PORTN_DIR_R        (*((volatile uint32_t *)0x40064400))
#define GPIO_PORTN_AFSEL_R      (*((volatile uint32_t *)0x40064420))
#define GPIO_PORTN_DEN_R        (*((volatile uint32_t *)0x4006451C))
#define GPIO_PORTN_AMSEL_R      (*((volatile uint32_t *)0x40064528))
#define GPIO_PORTN_PCTL_R       (*((volatile uint32_t *)0x4006452C))
#define SYSCTL_RCGCGPIO_R       (*((volatile uint32_t *)0x400FE608))
#define SYSCTL_RCGCGPIO_R12     0x00001000  // GPIO Port N Run Mode Clock
                                            // Gating Control
#define SYSCTL_PRGPIO_R         (*((volatile uint32_t *)0x400FEA08))
#define SYSCTL_PRGPIO_R12       0x00001000  // GPIO Port N Peripheral Ready
#define NVIC_ST_CTRL_R          (*((volatile uint32_t *)0xE000E010))
#define NVIC_ST_CTRL_COUNT      0x00010000  // Count flag
#define NVIC_ST_CTRL_CLK_SRC    0x00000004  // Clock Source
#define NVIC_ST_CTRL_INTEN      0x00000002  // Interrupt enable
#define NVIC_ST_CTRL_ENABLE     0x00000001  // Counter mode
#define NVIC_ST_RELOAD_R        (*((volatile uint32_t *)0xE000E014))
#define NVIC_ST_RELOAD_M        0x00FFFFFF  // Counter load value
#define NVIC_ST_CURRENT_R       (*((volatile uint32_t *)0xE000E018))

// Initialize SysTick with busy wait running at bus clock.
void SysTick_Init(void){
  NVIC_ST_CTRL_R = 0;                   // disable SysTick during setup
  NVIC_ST_RELOAD_R = NVIC_ST_RELOAD_M;  // maximum reload value
  NVIC_ST_CURRENT_R = 0;                // any write to current clears it
                                        // enable SysTick with core clock
  NVIC_ST_CTRL_R = NVIC_ST_CTRL_ENABLE+NVIC_ST_CTRL_CLK_SRC;
}

#define ASSEMBLY 0
#define RECURSIVE 0

#if ((ASSEMBLY == 0) && (RECURSIVE == 0))
  // iterative implementation of factorial
  // Assumes: n <= 12
  uint32_t Fact(uint32_t n){
  uint32_t r;
    r = 1;
    for(; n>1; n--){
      r = r*n;
    }
    return r;
  }
#else
  #if ((ASSEMBLY == 0) && (RECURSIVE != 0))
    // recursive implementation of factorial
    // Assumes: n <= 12
    uint32_t Fact(uint32_t n){
      if(n <= 1){   // end condition
        return 1;
      }
      return n*Fact(n-1); // recursion
    }
  #else
    #if ((ASSEMBLY != 0) && (RECURSIVE == 0))
      // iterative implementation of factorial (22 bytes)
      // Input: R0 is n
      // Output: R0 is Fact(n)
      // Assumes: R0 <= 12 (13! overflows)
      #ifdef __TI_COMPILER_VERSION__
          //Code Composer Studio Code
          uint32_t Fact(uint32_t n){
          __asm (  "     MOV R1, #1  ; R1 = 1 = total\n"
              "loop CMP R0, #1  ; is n (R0) <= 1?\n"
              "     BLS done    ; if so, skip to done\n"
              "     MUL R1, R0, R1  ; total = total*n\n"
              "     SUB R0, R0, #1  ; n = n  1\n"
              "     B   loop\n"
              "done MOV R0, R1      ; total = Fact(n)\n"
              "     BX  LR\n");
        }
      #else
          //Keil uVision Code
          __asm uint32_t
          Fact(uint32_t n)
          {
     MOV R1, #1  ; R1 = 1 = total
loop CMP R0, #1  ; is n (R0) <= 1?
     BLS done    ; if so, skip to done
     MUL R1, R0, R1  ; total = total*n
     SUB R0, R0, #1  ; n = n - 1
     B   loop
done MOV R0, R1      ; total = Fact(n)
             BX  LR
          }
      #endif
    #else
      #if ((ASSEMBLY != 0) && (RECURSIVE != 0))
        // recursive implementation of factorial (30 bytes)
        // Input: R0 is n
        // Output: R0 is Fact(n)
        // Assumes: R0 <= 12 (13! overflows)
        #ifdef __TI_COMPILER_VERSION__
            //Code Composer Studio Code
            uint32_t Fact(uint32_t n){
            __asm (  "     CMP R0, #1    ; is n (R0) <= 1?\n"
                "     BLS endcase   ; if so, to endcase\n"
                "     PUSH {R0, LR}  ; save R0 and LR\n"
                "     SUB R0, R0, #1 ; n = n - 1\n"
                "     BL  Fact       ; R0 = Fact(n-1)\n"
                "     POP {R1, LR}   ; restore R1, LR\n"
                "     MUL R0, R0, R1 ; R0 = n*Fact(n-1)\n"
                "     BX  LR         ; normal return\n"
                "endcase\n"
                "     MOV R0, #1     ; R0 = 1\n"
                "     BX  LR         ; end case return\n");
          }
        #else
            //Keil uVision Code
            __asm uint32_t
            Fact(uint32_t n)
            {
     CMP R0, #1    ; is n (R0) <= 1?
     BLS endcase   ; if so, to endcase
     PUSH {R0, LR}  ; save R0 and LR
     SUB R0, R0, #1 ; n = n - 1
     BL  Fact       ; R0 = Fact(n-1)
     POP {R1, LR}   ; restore R1, LR
     MUL R0, R0, R1 ; R0 = n*Fact(n-1)
     BX  LR         ; normal return
endcase
     MOV R0, #1     ; R0 = 1
     BX  LR         ; end case return
            }
        #endif
      #endif
    #endif
  #endif
#endif

uint32_t before,elapsed,n,f;
int main(void){
  PLL_Init();                      // set system clock to 120 MHz
  SysTick_Init();                  // initialize SysTick timer, Program 4.7
                                   // activate clock for Port N
  SYSCTL_RCGCGPIO_R |= SYSCTL_RCGCGPIO_R12;
                                   // allow time for clock to stabilize
  while((SYSCTL_PRGPIO_R&SYSCTL_PRGPIO_R12) == 0){};
  GPIO_PORTN_DIR_R |= 0x06;        // make PN2-1 out (PN1 built-in LED1, PN2 easier to reach)
  GPIO_PORTN_AFSEL_R &= ~0x06;     // disable alt funct on PN2-1
  GPIO_PORTN_DEN_R |= 0x06;        // enable digital I/O on PN2-1
                                   // configure PN2-1 as GPIO
  GPIO_PORTN_PCTL_R = (GPIO_PORTN_PCTL_R&0xFFFFF00F)+0x00000000;
  GPIO_PORTN_AMSEL_R &= ~0x06;     // disable analog functionality on PN2-1
  n = 1;
  before = NVIC_ST_CURRENT_R;
  f = Fact(n);
  elapsed = (before-NVIC_ST_CURRENT_R-15)&0x00FFFFFF;

  while(1){
    n = 1;
    GPIO_PORTN21 = 0x06;           //  PN2-1=1
    f = Fact(n);
    GPIO_PORTN21 = 0x00;           //  PN2-1=0
  }
}
